package dtu.library.acceptance_tests;

import static org.hamcrest.CoreMatchers.*;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertThat;
import static org.junit.Assert.assertTrue;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import cucumber.api.java.en.Given;
import cucumber.api.java.en.Then;
import cucumber.api.java.en.When;
import dtu.library.app.Book;
import dtu.library.app.LibraryApp;
import dtu.library.app.OperationNotAllowedException;
import dtu.library.app.TooManyBooksException;

public class BookSteps {

	private LibraryApp libraryApp;

	private Book book;
	private ErrorMessageHolder errorMessage;
	private List<Book> books;
	
	private UserHelper helper;

	/*
	 * Note that the constructor is apparently never called, but there are no null
	 * pointer exceptions regarding that libraryApp is not set. When creating the
	 * BookSteps object, the Cucumber libraries are using that constructor with an
	 * object of class LibraryApp as the default.
	 * 
	 * This also holds for all other step classes that have a similar constructor.
	 * In this case, the <b>same</b> object of class LibraryApp is used as an
	 * argument. This provides an easy way of sharing the same object, in this case
	 * the object of class LibraryApp, among all step classes.
	 * 
	 * This principle is called <em>dependency injection</em>. More information can
	 * be found in the "Cucumber for Java" book available online from the DTU Library.
	 */
	public BookSteps(LibraryApp libraryApp, ErrorMessageHolder errorMessage, UserHelper helper) {
		this.libraryApp = libraryApp;
		this.errorMessage = errorMessage;
		this.helper = helper;
	}

	@Given("^I have a book with title \"([^\"]*)\", author \"([^\"]*)\", and signature \"([^\"]*)\"$")
	public void iHaveABookWithTitleAuthorAndSignature(String title, String author, String signature) throws Exception {
		book = new Book(title,author,signature);
	}

	@Given("^a book with signature \"([^\"]*)\" is in the library$")
	public void aBookWithSignatureIsInTheLibrary(String signature) throws Exception {
		book = new Book("Extreme Programming", "Kent Beck", signature);
		addBooksToLibrary(Arrays.asList(book));
	}

	@Given("^these books are contained in the library$")
	public void theseBooksAreContainedInTheLibrary(List<List<String>> books) throws Exception {
		for (List<String> bookInfo : books) {
			libraryApp.addBook(new Book(bookInfo.get(0), bookInfo.get(1), bookInfo.get(2)));
		}
	}
	
	@When("^I add the book$")
	public void iAddTheBook() throws Exception {
		try {
			libraryApp.addBook(book);
		} catch (OperationNotAllowedException e) {
			errorMessage.setErrorMessage(e.getMessage());
		}
	}

	@Then("^the book with title \"([^\"]*)\", author \"([^\"]*)\", and signature \"([^\"]*)\" is added to the library$")
	public void theBookWithTitleAuthorAndSignatureIsAddedToTheLibrary(String title, String author, String signature)
			throws Exception {
		assertThat(book.getTitle(),is(equalTo(title)));
		assertThat(book.getAuthor(),is(equalTo(author)));
		assertThat(book.getSignature(),is(equalTo(signature)));
		assertThat(libraryApp.getBooks(),hasItem(book));
	}

	@Then("^I get the error message \"([^\"]*)\"$")
	public void iGetTheErrorMessage(String errorMessage) throws Exception {
		assertEquals(errorMessage, this.errorMessage.getErrorMessage());
	}

	@Given("^the library has a book with title \"([^\"]*)\", author \"([^\"]*)\", and signature \"([^\"]*)\"$")
	public void theLibraryHasABookWithTitleAuthorAndSignature(String title, String author, String signature)
			throws Exception {
		Book book = new Book(title, author, signature);
		assertThat(book.getTitle(),is(equalTo(title)));
		assertThat(book.getAuthor(),is(equalTo(author)));
		assertThat(book.getSignature(),is(equalTo(signature)));
		libraryApp.addBook(book);
	}

	@When("^I search for the text \"([^\"]*)\"$")
	public void iSearchForTheText(String searchText) throws Exception {
		books = libraryApp.search(searchText);
	}

	@Then("^I find the book with signature \"([^\"]*)\"$")
	public void iFindTheBookWithSignature(String signature) throws Exception {
		assertEquals(1, books.size());
		assertEquals(signature, books.get(0).getSignature());
	}

	@Then("^I don't find any book$")
	public void iDonTFindAnyBook() throws Exception {
		assertTrue(books.isEmpty());
	}

	@Then("^I find a book with signatures \"([^\"]*)\" and \"([^\"]*)\"$")
	public void iFindABookWithSignaturesAnd(String signature1, String signature2) throws Exception {
		assertEquals(2, books.size());
		Book book1 = books.get(0);
		Book book2 = books.get(1);
		assertTrue((book1.getSignature().equals(signature1) && book2.getSignature().equals(signature2))
				|| (book1.getSignature().equals(signature2) && book2.getSignature().equals(signature1)));
	}
	
	@Given("^book \"([^\"]*)\" by \"([^\"]*)\" with signature \"([^\"]*)\" is in the library$")
	public void bookByWithSignatureIsInTheLibrary(String title, String author, String signature) throws Exception {
		book = new Book(title,author,signature);
	}
	
	@Given("^the user has borrowed (\\d+) books$")
	public void theUserHasBorrowedBooks(int arg1) throws Exception {
		List<Book> exampleBooks = getExampleBooks(10);
		addBooksToLibrary(exampleBooks);
		for (Book b : exampleBooks) {
			helper.getUser().borrowBook(b);
		}
	}

	private void addBooksToLibrary(List<Book> exampleBooks) throws OperationNotAllowedException {
		libraryApp.adminLogin("adminadmin");
		for (Book b : exampleBooks) {
			libraryApp.addBook(b);
		}
		libraryApp.adminLogout();
	}

	private List<Book> getExampleBooks(int n) {
		List<Book> books = new ArrayList<>();
		for (int i = 1; i <= n; i++) {
			books.add(new Book("title"+i,"author"+i,"signature"+i));
		}
		return books;
	}

	@When("^the user borrows the book$")
	public void theUserBorrowsTheBook() throws Exception {
		try {
			helper.getUser().borrowBook(book);
		} catch (TooManyBooksException e ) {
			errorMessage.setErrorMessage(e.getMessage());
		}
	}

	@Then("^the book is borrowed by the user$")
	public void theBookIsBorrwedByTheUser() throws Exception {
		assertThat(helper.getUser().getBorrowedBooks(),hasItem(book));
	}
	
	@Then("^the book is not borrowed by the user$")
	public void theBookIsNotBorrowedByTheUser() throws Exception {
		assertThat(helper.getUser().getBorrowedBooks(),not(hasItem(book)));
	}
	
	@Given("^that the user has borrowed a book$")
	public void thatTheUserHasBorrowedABook() throws Exception {
		book = new Book("title","author","signature");
		addBooksToLibrary(Arrays.asList(book));
		helper.getUser().borrowBook(book);
	}
	
	@Given("^that the user not borrowed the book$")
	public void thatTheUserNotBorrowedTheBook() throws Exception {
		book = new Book("title","author","signature");
		addBooksToLibrary(Arrays.asList(book));
	}

	@When("^the user returns the book$")
	public void theUserReturnsTheBook() throws Exception {
		try {
			helper.getUser().returnBook(book);
		} catch (Exception e) {
			errorMessage.setErrorMessage(e.getMessage());
		}
	}
}
